/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#include "log_manager.h"

#include <cstring>
#include <mutex>

/**
 * @description: 添加日志记录到日志缓冲区中，并返回日志记录号
 * @param {LogRecord*} log_record 要写入缓冲区的日志记录
 * @return {lsn_t} 返回该日志的日志记录号
 */
lsn_t LogManager::add_log_to_buffer(LogRecord* log_record) {
  std::unique_lock<std::mutex> lock(latch_);

  if (log_buffer_.is_full(log_record->log_tot_len_)) {
    flush_log_to_disk();
    cv_.notify_all();
  }
  log_record->lsn_ = ++global_lsn_;
  log_record->serialize(log_buffer_.buffer_ + log_buffer_.offset_);
  log_buffer_.offset_ += log_record->log_tot_len_;

  return global_lsn_;
}

/**
 * @description: 把日志缓冲区的内容刷到磁盘中，由于目前只设置了一个缓冲区，因此需要阻塞其他日志操作
 */
void LogManager::flush_log_to_disk() {
  disk_manager_->write_log(log_buffer_.buffer_, log_buffer_.offset_);
  log_buffer_.offset_ = 0;

  int t = global_lsn_;
  persist_lsn_ = t;

  int fd = disk_manager_->open_file(LOG_MASTER_FILE_NAME);
  lsn_t persist_lsn = get_flushed_lsn();
  disk_manager_->write_page(fd, HEADER_PAGE_ID, (const char*)&persist_lsn, sizeof(lsn_t));
  disk_manager_->close_file(fd);
}
